#!/usr/bin/env python3
# this module is part of undetected_chromedriver

from distutils.version import LooseVersion
import io
import logging
import os
import pathlib
import random
import re
import shutil
import string
import subprocess
import sys
import time
from urllib.request import urlopen
from urllib.request import urlretrieve
import zipfile
from multiprocessing import Lock

logger = logging.getLogger(__name__)

IS_POSIX = sys.platform.startswith(("darwin", "cygwin", "linux", "linux2"))


class Patcher(object):
    lock = Lock()
    url_repo = "https://chromedriver.storage.googleapis.com"
    zip_name = "chromedriver_%s.zip"
    exe_name = "chromedriver%s"

    platform = sys.platform
    if platform.endswith("win32"):
        zip_name %= "win32"
        exe_name %= ".exe"
    if platform.endswith(("linux", "linux2")):
        zip_name %= "linux64"
        exe_name %= ""
    if platform.endswith("darwin"):
        zip_name %= "mac64"
        exe_name %= ""

    if platform.endswith("win32"):
        d = "~/appdata/roaming/undetected_chromedriver"
    elif "LAMBDA_TASK_ROOT" in os.environ:
        d = "/tmp/undetected_chromedriver"
    elif platform.startswith(("linux", "linux2")):
        d = "~/.local/share/undetected_chromedriver"
    elif platform.endswith("darwin"):
        d = "~/Library/Application Support/undetected_chromedriver"
    else:
        d = "~/.undetected_chromedriver"
    data_path = os.path.abspath(os.path.expanduser(d))

    def __init__(
        self,
        executable_path=None,
        force=False,
        version_main: int = 0,
        user_multi_procs=False,
    ):
        """
        Args:
            executable_path: None = automatic
                             a full file path to the chromedriver executable
            force: False
                    terminate processes which are holding lock
            version_main: 0 = auto
                specify main chrome version (rounded, ex: 82)
        """
        self.force = force
        self._custom_exe_path = False
        prefix = "undetected"
        self.user_multi_procs = user_multi_procs

        if not os.path.exists(self.data_path):
            os.makedirs(self.data_path, exist_ok=True)

        if not executable_path:
            self.executable_path = os.path.join(
                self.data_path, "_".join([prefix, self.exe_name])
            )

        if not IS_POSIX:
            if executable_path:
                if not executable_path[-4:] == ".exe":
                    executable_path += ".exe"

        self.zip_path = os.path.join(self.data_path, prefix)

        if not executable_path:
            if not self.user_multi_procs:
                self.executable_path = os.path.abspath(
                    os.path.join(".", self.executable_path)
                )

        if executable_path:
            self._custom_exe_path = True
            self.executable_path = executable_path

        self.version_main = version_main
        self.version_full = None

    def auto(self, executable_path=None, force=False, version_main=None, _=None):
        """

        Args:
            executable_path:
            force:
            version_main:

        Returns:

        """
        # if self.user_multi_procs and \
        #         self.user_multi_procs != -1:
        #     # -1 being a skip value used later in this block
        #
        p = pathlib.Path(self.data_path)
        # with Lock():
        #     files = list(p.rglob("*chromedriver*?"))
        #     for file in files:
        #         if self.is_binary_patched(file):
        #             self.executable_path = str(file)
        #             return True

        if executable_path:
            self.executable_path = executable_path
            self._custom_exe_path = True

        if self._custom_exe_path:
            file_name, file_extension = os.path.splitext(self.executable_path)
            # 创建新的文件名
            new_file = f"{file_name}_uc{file_extension}"
            if not os.path.exists(new_file):
                shutil.copy(self.executable_path, new_file)
            self.executable_path = new_file # 用新的chromedriver
            print(f"New chromedriver path: {self.executable_path}")
            ispatched = self.is_binary_patched(self.executable_path)
            folder_path = os.path.dirname(os.path.abspath(self.executable_path))
            folder_list = [f for f in os.listdir(folder_path) if os.path.isdir(os.path.join(folder_path, f))]
            numeric_folders = [f for f in folder_list if f[0].isdigit()]
            version = numeric_folders[0].split('.')[0]
            # print(f"\n\n\nCloudflare下需要自行安装浏览器，请确保自己的机器环境已经安装了 {numeric_folders[0].split('.')[0]} 版本的Chrome浏览器（不是软件自带的Chrome浏览器，需要自己安装浏览器且版本号一定要正确），否则程序无法运行！")
            # print("Please make sure that your machine environment has installed the Chrome browser version %s (not the Chrome browser provided by the software, you need to install the browser yourself and the version number must be correct), otherwise the program cannot run!" % numeric_folders[0].split('.')[0])

            if not ispatched:
                print("Patching chromedriver...")
                return version
            else:
                print("No need to patch chromedriver.")
                return version

        if version_main:
            self.version_main = version_main
        if force is True:
            self.force = force

        try:
            os.unlink(self.executable_path)
        except PermissionError:
            if self.force:
                self.force_kill_instances(self.executable_path)
                return self.auto(force=not self.force)
            try:
                if self.is_binary_patched():
                    # assumes already running AND patched
                    return True
            except PermissionError:
                pass
            # return False
        except FileNotFoundError:
            pass

        release = self.fetch_release_number()
        self.version_main = release.version[0]
        self.version_full = release
        self.unzip_package(self.fetch_package())
        return self.patch()

    def driver_binary_in_use(self, path: str = None) -> bool:
        """
        naive test to check if a found chromedriver binary is
        currently in use

        Args:
            path: a string or PathLike object to the binary to check.
                  if not specified, we check use this object's executable_path
        """
        if not path:
            path = self.executable_path
        p = pathlib.Path(path)

        if not p.exists():
            raise OSError("file does not exist: %s" % p)
        try:
            with open(p, mode="a+b") as fs:
                exc = []
                try:

                    fs.seek(0, 0)
                except PermissionError as e:
                    exc.append(e)  # since some systems apprently allow seeking
                    # we conduct another test
                try:
                    fs.readline()
                except PermissionError as e:
                    exc.append(e)

                if exc:

                    return True
                return False
            # ok safe to assume this is in use
        except Exception as e:
            # logger.exception("whoops ", e)
            pass

    def cleanup_unused_files(self):
        p = pathlib.Path(self.data_path)
        items = list(p.glob("*undetected*"))
        print(items)

    def patch(self):
        self.patch_exe()
        return self.is_binary_patched()

    def fetch_release_number(self):
        """
        Gets the latest major version available, or the latest major version of self.target_version if set explicitly.
        :return: version string
        :rtype: LooseVersion
        """
        path = "/latest_release"
        if self.version_main:
            path += f"_{self.version_main}"
        path = path.upper()
        logger.debug("getting release number from %s" % path)
        return LooseVersion(urlopen(self.url_repo + path).read().decode())

    def parse_exe_version(self):
        with io.open(self.executable_path, "rb") as f:
            for line in iter(lambda: f.readline(), b""):
                match = re.search(rb"platform_handle\x00content\x00([0-9.]*)", line)
                if match:
                    return LooseVersion(match[1].decode())

    def fetch_package(self):
        """
        Downloads ChromeDriver from source

        :return: path to downloaded file
        """
        u = "%s/%s/%s" % (self.url_repo, self.version_full.vstring, self.zip_name)
        logger.debug("downloading from %s" % u)
        # return urlretrieve(u, filename=self.data_path)[0]
        return urlretrieve(u)[0]

    def unzip_package(self, fp):
        """
        Does what it says

        :return: path to unpacked executable
        """
        logger.debug("unzipping %s" % fp)
        try:
            os.unlink(self.zip_path)
        except (FileNotFoundError, OSError):
            pass

        os.makedirs(self.zip_path, mode=0o755, exist_ok=True)
        with zipfile.ZipFile(fp, mode="r") as zf:
            zf.extract(self.exe_name, self.zip_path)
        os.rename(os.path.join(self.zip_path, self.exe_name), self.executable_path)
        os.remove(fp)
        os.rmdir(self.zip_path)
        os.chmod(self.executable_path, 0o755)
        return self.executable_path

    @staticmethod
    def force_kill_instances(exe_name):
        """
        kills running instances.
        :param: executable name to kill, may be a path as well

        :return: True on success else False
        """
        exe_name = os.path.basename(exe_name)
        if IS_POSIX:
            r = os.system("kill -f -9 $(pidof %s)" % exe_name)
        else:
            r = os.system("taskkill /f /im %s" % exe_name)
        return not r

    @staticmethod
    def gen_random_cdc():
        cdc = random.choices(string.ascii_letters, k=27)
        return "".join(cdc).encode()

    def is_binary_patched(self, executable_path=None):
        executable_path = executable_path or self.executable_path
        try:
            with io.open(executable_path, "rb") as fh:
                return fh.read().find(b"undetected chromedriver") != -1
        except FileNotFoundError:
            return False

    def patch_exe(self):
        start = time.perf_counter()
        print("patching driver executable %s" % self.executable_path)
        with io.open(self.executable_path, "r+b") as fh:
            content = fh.read()
            # match_injected_codeblock = re.search(rb"{window.*;}", content)
            match_injected_codeblock = re.search(rb"\{window\.cdc.*?;\}", content)
            if match_injected_codeblock:
                target_bytes = match_injected_codeblock[0]
                new_target_bytes = (
                    b'{console.log("undetected chromedriver 1337!")}'.ljust(
                        len(target_bytes), b" "
                    )
                )
                new_content = content.replace(target_bytes, new_target_bytes)
                if new_content == content:
                    logger.warning(
                        "something went wrong patching the driver binary. could not find injection code block"
                    )
                else:
                    logger.debug(
                        "found block:\n%s\nreplacing with:\n%s"
                        % (target_bytes, new_target_bytes)
                    )
                fh.seek(0)
                fh.write(new_content)
        logger.debug(
            "patching took us {:.2f} seconds".format(time.perf_counter() - start)
        )

    def __repr__(self):
        return "{0:s}({1:s})".format(
            self.__class__.__name__,
            self.executable_path,
        )

    def __del__(self):
        if self._custom_exe_path:
            # if the driver binary is specified by user
            # we assume it is important enough to not delete it
            return
        else:
            timeout = 3  # stop trying after this many seconds
            t = time.monotonic()
            now = lambda: time.monotonic()
            while now() - t > timeout:
                # we don't want to wait until the end of time
                try:
                    if self.user_multi_procs:
                        break
                    os.unlink(self.executable_path)
                    logger.debug("successfully unlinked %s" % self.executable_path)
                    break
                except (OSError, RuntimeError, PermissionError):
                    time.sleep(0.01)
                    continue
                except FileNotFoundError:
                    break



import logging
from pathlib import Path

from mtmai_client.client import Client

from mtmtrain.core.config import settings

logger = logging.getLogger()

local_dataset_dir = "datasets"


def download_dataset(dataset: str):
    from mtmai_client.api.dataset import dataset_dataset_download

    client = Client(base_url=settings.MTMAI_API_BASE)

    # for ds in all_datasets:
    response = dataset_dataset_download.sync(client=client, dataset_path=dataset)
    dataset_path = get_dataset_path(dataset)
    dataset_path.parent.mkdir(parents=True, exist_ok=True)
    dataset_path.write_text(response)
    logger.info("数据集下载完成 %s", dataset_path.resolve())


def down_dataset_from_url(url: str):
    localbase_dir = Path(settings.storage_dir).joinpath(local_dataset_dir)
    # 输入的网址形如： https://cdn-datasets.huggingface.co/EsperBERTo/data/oscar.eo.txt

    # 使用 httpx 下载文件，并且将文件保存到 f{localbase_dir}/oscar.eo.txt 下。


def load_dataset(dataset_name: str):
    local_path = (
        Path(settings.storage_dir).joinpath(local_dataset_dir).joinpath(dataset_name)
    )
    if not local_path.exists():
        download_dataset(dataset_name)


def get_dataset_path(dataset: str):
    # settings.storage_dir 值为 ".vol", local_dataset_dir= "datasets"
    # settings.storage_dir 值为 ".vol", local_dataset_dir= "datasets", ="common/Tweets.csv"
    a = Path(settings.storage_dir).joinpath(local_dataset_dir).joinpath(dataset)
    print(str(a))
    return a



from typing import Any, Type

from pydantic import BaseModel

from mtmai.deps import AsyncSessionDep, CurrentUser


class FlowBase:
    @classmethod
    async def execute(cls) -> Any:
        raise NotImplementedError("Subclasses must implement this method")


def mtflow(form_model: Type[BaseModel]):
    def decorator(cls):
        cls.form_model = form_model
        return cls

    return decorator



import asyncio
from collections.abc import Mapping
from collections.abc import Sequence
from functools import wraps
import logging
import threading
import time
import traceback
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import List
from typing import Optional


class Structure(dict):
    """
    This is a dict-like object structure, which you should subclass
    Only properties defined in the class context are used on initialization.

    See example
    """

    _store = {}

    def __init__(self, *a, **kw):
        """
        Instantiate a new instance.

        :param a:
        :param kw:
        """

        super().__init__()

        # auxiliar dict
        d = dict(*a, **kw)
        for k, v in d.items():
            if isinstance(v, Mapping):
                self[k] = self.__class__(v)
            elif isinstance(v, Sequence) and not isinstance(v, (str, bytes)):
                self[k] = [self.__class__(i) for i in v]
            else:
                self[k] = v
        super().__setattr__("__dict__", self)

    def __getattr__(self, item):
        return getattr(super(), item)

    def __getitem__(self, item):
        return super().__getitem__(item)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def update(self, *a, **kw):
        super().update(*a, **kw)

    def __eq__(self, other):
        return frozenset(other.items()) == frozenset(self.items())

    def __hash__(self):
        return hash(frozenset(self.items()))

    @classmethod
    def __init_subclass__(cls, **kwargs):
        cls._store = {}

    def _normalize_strings(self):
        for k, v in self.copy().items():
            if isinstance(v, (str)):
                self[k] = v.strip()


def timeout(seconds=3, on_timeout: Optional[Callable[[callable], Any]] = None):
    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            def function_reached_timeout():
                if on_timeout:
                    on_timeout(func)
                else:
                    raise TimeoutError("function call timed out")

            t = threading.Timer(interval=seconds, function=function_reached_timeout)
            t.start()
            try:
                return func(*args, **kwargs)
            except:
                t.cancel()
                raise
            finally:
                t.cancel()

        return wrapped

    return wrapper


def test():
    import sys, os

    sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
    import undetected_chromedriver as uc
    import threading

    def collector(
        driver: uc.Chrome,
        stop_event: threading.Event,
        on_event_coro: Optional[Callable[[List[str]], Awaitable[Any]]] = None,
        listen_events: Sequence = ("browser", "network", "performance"),
    ):
        def threaded(driver, stop_event, on_event_coro):
            async def _ensure_service_started():
                while (
                    getattr(driver, "service", False)
                    and getattr(driver.service, "process", False)
                    and driver.service.process.poll()
                ):
                    print("waiting for driver service to come back on")
                    await asyncio.sleep(0.05)
                    # await asyncio.sleep(driver._delay or .25)

            async def get_log_lines(typ):
                await _ensure_service_started()
                return driver.get_log(typ)

            async def looper():
                while not stop_event.is_set():
                    log_lines = []
                    try:
                        for _ in listen_events:
                            try:
                                log_lines += await get_log_lines(_)
                            except:
                                if logging.getLogger().getEffectiveLevel() <= 10:
                                    traceback.print_exc()
                                continue
                        if log_lines and on_event_coro:
                            await on_event_coro(log_lines)
                    except Exception as e:
                        if logging.getLogger().getEffectiveLevel() <= 10:
                            traceback.print_exc()

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(looper())

        t = threading.Thread(target=threaded, args=(driver, stop_event, on_event_coro))
        t.start()

    async def on_event(data):
        print("on_event")
        print("data:", data)

    def func_called(fn):
        def wrapped(*args, **kwargs):
            print(
                "func called! %s  (args: %s, kwargs: %s)" % (fn.__name__, args, kwargs)
            )
            while driver.service.process and driver.service.process.poll() is not None:
                time.sleep(0.1)
            res = fn(*args, **kwargs)
            print("func completed! (result: %s)" % res)
            return res

        return wrapped

    logging.basicConfig(level=10)

    options = uc.ChromeOptions()
    options.set_capability(
        "goog:loggingPrefs", {"performance": "ALL", "browser": "ALL", "network": "ALL"}
    )

    driver = uc.Chrome(version_main=96, options=options)

    # driver.command_executor._request = timeout(seconds=1)(driver.command_executor._request)
    driver.command_executor._request = func_called(driver.command_executor._request)
    collector_stop = threading.Event()
    collector(driver, collector_stop, on_event)

    driver.get("https://nowsecure.nl")

    time.sleep(10)

    driver.quit()



import yaml


def load_yaml_file(filename, label='config file', loader=yaml.Loader):
    """Load a yaml config file.
    """
    try:
        with open(filename, "r") as file:
            res = yaml.load(file.read(), Loader=loader) or {}
            if not isinstance(res, dict):
                msg = f"Please ensure that {label} consists of key value pairs."
                raise ValueError(f"Invalid {label}: {filename}", msg)
            return res

    except yaml.YAMLError as err:
        if hasattr(err, 'problem_mark'):
            mark = getattr(err, 'problem_mark')
            problem = getattr(err, 'problem')
            message = f"Could not read {label} {filename}:"
            message += "\n filename: {filename}, mark line: {mark.line}, column: {mark.column} {problem}"
        elif hasattr(err, 'problem'):
            problem = getattr(err, 'problem')
            message = f"Could not read {label} {filename}: {problem}"
        else:
            message = f"Could not read {label} {filename}: YAML Error"

        suggestion = f"There is a syntax error in your {label} - please fix it and try again."

        raise ValueError(message, suggestion)

    except OSError as err:
        msg = "Please ensure the file exists and you have the required access privileges."
        raise ValueError(f"Could not open {label} {filename}: {err.strerror}", msg)



def save_dict_to_yaml(dict_value: dict, save_path: str):
    """dict保存为yaml"""
    with open(save_path, 'w') as file:
        file.write(yaml.dump(dict_value, allow_unicode=True))
    with open(save_path, 'w') as file:
        file.write(yaml.dump(dict_value, allow_unicode=True))



import structlog
from fastapi import APIRouter, Depends
from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_oauth2_redirect_html

# from mtmlib.decorators.mtform.mtform import MtForm
from opentelemetry import trace
from pydantic import BaseModel

from mtmai.agents.opencanvas.opencanvas_state import OpenCanvasState
from mtmai.core.config import settings
from mtmai.deps import get_current_active_superuser
from mtmai.models.agent import ChatBotUiStateResponse
from mtmai.models.chat import ChatProfile, ListViewProps, ThreadUIState

# from mtmai.models.models import Message
from mtmai.mtlibs.inputs.input_widget import ThreadForm

# from mtmai.types.types import Theme
# from mtmai.utils import generate_test_email, send_email

tracer = trace.get_tracer_provider().get_tracer(__name__)
LOG = structlog.get_logger()


# mimetypes.add_type("application/javascript", ".js")
# mimetypes.add_type("text/css", ".css")


router = APIRouter()


@router.get("/health", include_in_schema=False)
async def health_check():
    with tracer.start_as_current_span("health-span"):
        LOG.info("get /health")
        current_span = trace.get_current_span()
        current_span.add_event("This is a span event")
        LOG.warning("This is a log message")
        return {"health": True}


@router.get("/swagger-ui-oauth2-redirect", include_in_schema=False)
async def swagger_ui_redirect():
    return get_swagger_ui_oauth2_redirect_html()


@router.get("/redoc", include_in_schema=False)
async def redoc_html():
    return get_redoc_html(
        openapi_url=router.app().openapi_url,
        title=router.app().title + " - ReDoc",
        redoc_js_url="https://unpkg.com/redoc@next/bundles/redoc.standalone.js",
    )


@router.get("/info", include_in_schema=False)
async def app_info():
    return {
        "app_name": settings.app_name,
        "admin_email": settings.admin_email,
        "items_per_user": settings.items_per_user,
    }


@router.post(
    "/test-email/",
    dependencies=[Depends(get_current_active_superuser)],
    status_code=201,
    include_in_schema=False,
)
class TestUrlReq(BaseModel):
    url: str


class TypesResponse(BaseModel):
    """
    如果使用openapi 生成前端代码，缺少了某些类型，请在这里补充
    """

    thread_form: ThreadForm | None = None
    uiState: ChatBotUiStateResponse | None = None
    thread_ui_state: ThreadUIState | None = None
    chat_profile: ChatProfile | None = None
    list_view_props: ListViewProps | None = None
    # mt_form: MtForm | None = None
    open_canvas_state: OpenCanvasState | None = None


@router.get("/types", include_in_schema=True, response_model=TypesResponse)
async def types():
    """
    无实际功能，仅用于openapi生成前端代码
    """
    return TypesResponse()


# @router.get("/logo")
# async def get_logo(theme: Optional[Theme] = Query(Theme.light)):
#     """Get the default logo for the UI."""
#     theme_value = theme.value if theme else Theme.light.value
#     logo_path = None

#     for path in [
#         os.path.join(APP_ROOT, "public", f"logo_{theme_value}.*"),
#         os.path.join(APP_ROOT, "mtmai/mtmai/public", f"logo_{theme_value}.*"),
#         # os.path.join(build_dir, "assets", f"logo_{theme_value}*.*"),
#     ]:
#         files = glob.glob(path)

#         if files:
#             logo_path = files[0]
#             break

#     if not logo_path:
#         raise HTTPException(status_code=404, detail="Missing default logo")

#     media_type, _ = mimetypes.guess_type(logo_path)

#     return FileResponse(logo_path, media_type=media_type)

#     # merge_gauge_data(filename)
#     # merge_gauge_data(filename)



import traceback
import uuid

import structlog
from fastapi.encoders import jsonable_encoder
from langchain_core.messages import HumanMessage
from langchain_core.runnables import RunnableConfig
from langgraph.graph import StateGraph
from langgraph.graph.graph import CompiledGraph
from langgraph.prebuilt import tools_condition
from langgraph.types import StateSnapshot

from mtmai.agents.ctx import init_mtmai_http_context, mtmai_context
from mtmai.agents.graphutils import (
    create_tool_node_with_fallback,
    is_internal_node,
    is_skip_kind,
    pop_dialog_state,
)
from mtmai.agents.task_graph.nodes.article_gen_node import ArticleGenNode
from mtmai.agents.task_graph.nodes.assisant_node import (
    PrimaryAssistantNode,
    primary_assistant_tools,
    route_assistant,
)
from mtmai.agents.task_graph.nodes.create_task_node import CreateTaskNode
from mtmai.agents.task_graph.nodes.human_input_node import HumanInputNode
from mtmai.agents.task_graph.nodes.site_node import SiteNode
from mtmai.agents.task_graph.nodes.task_entry_node import TaskEntryNode, routeEntryPath
from mtmai.agents.task_graph.task_state import TaskState
from mtmai.core.coreutils import is_in_dev
from mtmai.crud import curd_chat

LOG = structlog.get_logger()

HUMEN_INPUT_NODE = "human_input"


class TaskGraph:
    def __init__(self):
        pass

    @classmethod
    def name(cls):
        return "taskrunner"

    async def build_graph(self):
        wf = StateGraph(TaskState)

        wf.add_node("entry", TaskEntryNode())
        wf.set_entry_point("entry")
        wf.add_conditional_edges(
            "entry",
            routeEntryPath,
            [
                "articleGen",
                HUMEN_INPUT_NODE,
                "assistant",
                "site",
                "create_task",
            ],
        )

        wf.add_node("assistant", PrimaryAssistantNode())

        wf.add_conditional_edges(
            "assistant",
            tools_condition,
        )

        wf.add_node(
            "tools",
            create_tool_node_with_fallback(primary_assistant_tools),
        )
        wf.add_conditional_edges(
            "tools",
            route_assistant,
            {
                "assistant": "assistant",
                # "error": END,
            },
        )
        wf.add_node(HUMEN_INPUT_NODE, HumanInputNode())
        wf.add_edge(HUMEN_INPUT_NODE, "assistant")

        wf.add_node("articleGen", ArticleGenNode())
        wf.add_edge("articleGen", HUMEN_INPUT_NODE)

        wf.add_node("leave_skill", pop_dialog_state)
        wf.add_edge("leave_skill", "assistant")

        wf.add_node("site", SiteNode())
        wf.add_edge("site", "assistant")

        wf.add_node("create_task", CreateTaskNode())
        # wf.add_edge("create_task", "assistant")

        return wf

    async def compile_graph(self) -> CompiledGraph:
        graph = (await self.build_graph()).compile(
            checkpointer=await mtmai_context.get_graph_checkpointer(),
            # interrupt_after=["human"],
            interrupt_before=[
                HUMEN_INPUT_NODE,
                # "update_flight_sensitive_tools",
                # "develop_sensitive_tools",
                # "book_car_rental_sensitive_tools",
                # "book_hotel_sensitive_tools",
                # "book_excursion_sensitive_tools",
            ],
            debug=True,
        )

        if is_in_dev():
            image_data = graph.get_graph(xray=1).draw_mermaid_png()
            save_to = "./.vol/taskrunner_graph.png"
            with open(save_to, "wb") as f:
                f.write(image_data)
        return graph

    async def getExampleState(self):
        state = {
            "messages": [],
            "task_config": {
                "siteUrl": "https://www.baidu.com",
            },
        }
        return state

    async def chat_start(self):
        init_mtmai_http_context()
        user_session = cl.user_session
        user = user_session.get("user")
        thread_id = context.session.thread_id
        # await cl.Message(content="欢迎使用博客文章生成器").send()
        graph = await TaskGraph().compile_graph()
        user_session.set("graph", graph)

        context.session.has_first_interaction = True
        await context.emitter.emit(
            "first_interaction",
            {
                "interaction": "graph_start",
                "thread_id": context.session.thread_id,
            },
        )
        thread: RunnableConfig = {
            "configurable": {
                "thread_id": thread_id,
            }
        }
        # await self.run_graph(thread, {"messages": []})
        await self.run_graph(thread, await self.getExampleState())

    async def on_chat_resume(self):
        init_mtmai_http_context()
        thread_id = context.session.thread_id
        user_session = cl.user_session
        thread_id = context.session.thread_id
        await cl.Message(content="正在恢复对话").send()
        graph = await TaskGraph().compile_graph()
        user_session.set("graph", graph)

        context.session.has_first_interaction = True
        if not graph:
            cl.Message(content="工作流初始化失败").send()
            raise ValueError("graph 未初始化")

        thread: RunnableConfig = {
            "configurable": {
                "thread_id": thread_id,
            }
        }
        pre_state = await graph.aget_state(thread, subgraphs=True)
        LOG.info("()on_chat_resume %s", thread_id)

        all_steps = await curd_chat.get_steps_by_thread(thread_id)
        LOG.info("all_steps %s", all_steps)

    async def on_message(self, message: cl.Message):
        init_mtmai_http_context()
        try:
            user_session = cl.user_session
            thread_id = context.session.thread_id

            graph: CompiledGraph = user_session.get("graph")
            if not graph:
                cl.Message(content="工作流初始化失败").send()
                raise ValueError("graph 未初始化")

            thread: RunnableConfig = {
                "configurable": {
                    "thread_id": thread_id,
                    # 可以从指定检测点运行，以及分支
                    # "checkpoint_id": "xxxxx"
                }
            }

            all_checkpoints: list[StateSnapshot] = []
            async for state in graph.aget_state_history(thread):
                all_checkpoints.append(state)

            LOG.info("checkpoints: %d", len(all_checkpoints))

            pre_state: StateSnapshot = await graph.aget_state(thread, subgraphs=True)
            await graph.aupdate_state(
                thread,
                {
                    **pre_state.values,
                    "user_input": message.content,
                    "messages": [HumanMessage(content=message.content)],
                },
                as_node=HUMEN_INPUT_NODE,
            )
            await self.run_graph(thread)
        except Exception as e:
            error_message = f"An error occurred: {str(e)}\n\nDetailed traceback:\n{traceback.format_exc()}"
            LOG.error(error_message)
            await cl.Message(content=error_message).send()

    async def onRequest(self, inputs=None, messages=None, thread_id=None):
        graph = await TaskGraph().compile_graph()
        if not graph:
            raise ValueError("graph 未初始化")

        if not thread_id:
            thread_id = str(uuid.uuid4())

        thread: RunnableConfig = {
            "configurable": {
                "thread_id": thread_id,
                # 可以从指定检测点运行，以及分支
                # "checkpoint_id": "xxxxx"
            }
        }
        await self.run_graph(thread, graph, inputs, messages)

    async def run_graph(
        self, thread: RunnableConfig, graph: CompiledGraph, inputs=None, messages=None
    ):
        async for event in graph.astream_events(
            inputs,
            version="v2",
            config=thread,
            subgraphs=True,
        ):
            kind = event["event"]
            node_name = event["name"]
            data = event["data"]

            if not is_internal_node(node_name):
                if not is_skip_kind(kind):
                    LOG.info("[event] %s@%s", kind, node_name)
                    mtmai_context.emit("logs", {"on": kind, "node_name": node_name})

            if kind == "on_chat_model_stream":
                yield data

            if kind == "on_chain_start":
                LOG.info("on_chain_start %s:", node_name)
                output = data.get("output")
                if node_name == "__start__":
                    pass

            if kind == "on_chain_end":
                LOG.info("on_chain_end %s:", node_name)
                output = data.get("output")
                if node_name == "__start__":
                    pass
                if node_name in [HUMEN_INPUT_NODE, "articleGen", "entry"]:
                    human_ouput_message = output.get("human_ouput_message")
                    LOG.info("human_ouput_message %s", human_ouput_message)
            if node_name == "on_chat_start_node":
                thread_ui_state = output.get("thread_ui_state")
                if thread_ui_state:
                    await context.emitter.emit(
                        "ui_state_upate",
                        jsonable_encoder(thread_ui_state),
                    )

            if kind == "on_tool_start":
                await context.emitter.emit(
                    "logs",
                    {
                        "on": kind,
                        "node_name": node_name,
                    },
                )

            if kind == "on_tool_end":
                output = data.get("output")
                await context.emitter.emit(
                    "logs",
                    {
                        "on": kind,
                        "node_name": node_name,
                        "output": jsonable_encoder(output),
                    },
                )

    # 旧代码，跟 lainlit 集成，但是，使用 ws 作为事件发送，大大增加了复杂性。对于兼容性非常不友好。
    # async def run_graph(
    #     self,
    #     thread: RunnableConfig,
    #     inputs=None,
    # ):
    #     # ############################################################################################
    #     # TODO: 对于非流式 http 的处理方式：
    #     #     本质就是，在本地用一个对象内部 使用流式更新状态，知道 graph 解释，再response 发送到前端。
    #     ##############################################################################################
    #     user_session = cl.user_session
    #     graph = user_session.get("graph")
    #     if not graph:
    #         raise ValueError("graph 未初始化")

    #     # 流式传输过程：
    #     # 1. 先发送一个消息，让前端立即显示ai消息占位，后续流程处理可随时更新这个消息，包括流式传输。
    #     # 2. 一旦整个流程结束，再次调用 .send()，触发消息的持久化。

    #     resp_msg = None

    #     async def prepare_message():
    #         nonlocal resp_msg
    #         resp_msg = cl.Message(content="")
    #         await resp_msg.send()

    #     async def stream_token(a: str):
    #         nonlocal resp_msg
    #         if not a:
    #             return
    #         if not resp_msg:
    #             await prepare_message()
    #         await resp_msg.stream_token(a)

    #     async def stream_finish():
    #         nonlocal resp_msg
    #         if resp_msg:
    #             await resp_msg.send()

    #     async for event in graph.astream_events(
    #         inputs,
    #         version="v2",
    #         config=thread,
    #         subgraphs=True,
    #     ):
    #         kind = event["event"]
    #         node_name = event["name"]
    #         data = event["data"]

    #         if not is_internal_node(node_name):
    #             if not is_skip_kind(kind):
    #                 LOG.info("[event] %s@%s", kind, node_name)

    #         if kind == "on_chat_model_stream":
    #             content = event["data"]["chunk"].content
    #             await stream_token(content)

    #         # if kind == "on_chat_model_end":
    #         #     output = data.get("output")
    #         #     if output:
    #         #         chat_output = output.content
    #         #         await resp_msg.send()

    #         if kind == "on_chain_start":
    #             LOG.info("on_chain_start %s:", node_name)
    #             output = data.get("output")
    #             if node_name == "__start__":
    #                 pass

    #         if kind == "on_chain_end":
    #             LOG.info("on_chain_end %s:", node_name)
    #             output = data.get("output")
    #             if node_name == "__start__":
    #                 pass
    #             if node_name in [HUMEN_INPUT_NODE, "articleGen", "entry"]:
    #                 human_ouput_message = output.get("human_ouput_message")
    #                 if human_ouput_message:
    #                     await stream_token(human_ouput_message)
    #                     elements = [
    #                         cl.Text(content="element 测试1"),
    #                         cl.Text(content="element 测试2"),
    #                         cl.Text(
    #                             name="simple_text",
    #                             content="element 测试3",
    #                             display="inline",
    #                         ),
    #                         cl.Text(
    #                             name="点击侧边栏打开文本",
    #                             content="element 测试4,侧边栏显示",
    #                             display="side",
    #                         ),
    #                     ]
    #                     await cl.Message(content="element 测试1", elements=elements).send()

    #         if node_name == "on_chat_start_node":
    #             thread_ui_state = output.get("thread_ui_state")
    #             if thread_ui_state:
    #                 await context.emitter.emit(
    #                     "ui_state_upate",
    #                     jsonable_encoder(thread_ui_state),
    #                 )

    #         if kind == "on_tool_start":
    #             await context.emitter.emit(
    #                 "logs",
    #                 {
    #                     "on": kind,
    #                     "node_name": node_name,
    #                 },
    #             )

    #         if kind == "on_tool_end":
    #             output = data.get("output")
    #             await context.emitter.emit(
    #                 "logs",
    #                 {
    #                     "on": kind,
    #                     "node_name": node_name,
    #                     "output": jsonable_encoder(output),
    #                 },
    #             )
    #         if node_name == "LangGraph":
    #             LOG.info("中止节点")
    #             if (
    #                 data
    #                 and (output := data.get("output"))
    #                 and (final_messages := output.get("messages"))
    #             ):
    #                 for message in final_messages:
    #                     message.pretty_print()
    #                 await context.emitter.emit(
    #                     "logs",
    #                     {
    #                         "on": "pause",
    #                         "node_name": node_name,
    #                         "output": json.dumps(jsonable_encoder(message)),
    #                     },
    #                 )
    #             await stream_finish()



# from langchain_community.agent_toolkits import GmailToolkit
# from langchain_community.tools.gmail.get_thread import GmailGetThread
# from langchain_community.tools.tavily_search import TavilySearchResults

from textwrap import dedent

from crewai import Agent

from .tools import CreateDraftTool


class EmailFilterAgents:
    def __init__(self):
        # self.gmail = GmailToolkit()
        pass

    def email_filter_agent(self):
        return Agent(
            role="Senior Email Analyst",
            goal="Filter out non-essential emails like newsletters and promotional content",
            backstory=dedent("""\
				As a Senior Email Analyst, you have extensive experience in email content analysis.
				You are adept at distinguishing important emails from spam, newsletters, and other
				irrelevant content. Your expertise lies in identifying key patterns and markers that
				signify the importance of an email."""),
            verbose=True,
            allow_delegation=False,
        )

    def email_action_agent(self):
        return Agent(
            role="Email Action Specialist",
            goal="Identify action-required emails and compile a list of their IDs",
            backstory=dedent("""\
				With a keen eye for detail and a knack for understanding context, you specialize
				in identifying emails that require immediate action. Your skill set includes interpreting
				the urgency and importance of an email based on its content and context."""),
            tools=[
                # GmailGetThread(api_resource=self.gmail.api_resource),
                # TavilySearchResults(),
            ],
            verbose=True,
            allow_delegation=False,
        )

    def email_response_writer(self):
        return Agent(
            role="Email Response Writer",
            goal="Draft responses to action-required emails",
            backstory=dedent("""\
				You are a skilled writer, adept at crafting clear, concise, and effective email responses.
				Your strength lies in your ability to communicate effectively, ensuring that each response is
				tailored to address the specific needs and context of the email."""),
            tools=[
                # TavilySearchResults(),
                # GmailGetThread(api_resource=self.gmail.api_resource),
                CreateDraftTool.create_draft,
            ],
            verbose=True,
            allow_delegation=False,
        )



import asyncio
import json

from langchain_core.messages import AIMessage, ToolMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable, RunnableConfig
from pydantic import BaseModel, Field

from mtmai.agents.nodes.utils import swap_roles
from mtmai.agents.retrivers.web_search import search_engine
from mtmai.core.logging import get_logger
from mtmai.models.graph_config import InterviewState

logger = get_logger()


class AnswerWithCitations(BaseModel):
    answer: str = Field(
        description="Comprehensive answer to the user's question with citations.",
    )
    cited_urls: list[str] = Field(
        description="List of urls cited in the answer.",
    )

    @property
    def as_str(self) -> str:
        return f"{self.answer}\n\nCitations:\n\n" + "\n".join(
            f"[{i+1}]: {url}" for i, url in enumerate(self.cited_urls)
        )


class Queries(BaseModel):
    queries: list[str] = Field(
        description="Comprehensive list of search engine queries to answer the user's questions.",
    )


class GenAnswerNode:
    def __init__(
        self,
        runnable: Runnable,
        name: str = "Subject_Matter_Expert",
        max_str_len: int = 15000,
    ):
        self.runnable = runnable
        self.name = name
        self.max_str_len = max_str_len

    def node_name(self):
        return "gen_answer"

    async def __call__(
        self,
        state: InterviewState,
        config: RunnableConfig,
    ):
        logger.info(f"进入 gen_answer node: {state}")
        swapped_state = swap_roles(state, self.name)  # Convert all other AI messages

        # gen_queries_chain = gen_queries_prompt | self.runnable.with_structured_output(
        #     Queries, include_raw=True
        # ).with_retry(stop_after_attempt=3)

        # queries = await gen_queries_chain.ainvoke(swapped_state)
        queries = await self.gen_queries(swapped_state, self.runnable)
        parsed_queries = queries["parsed"]
        if not parsed_queries:
            logger.warning("生成搜索查询失败")
            return {"error": " gen_answer 没有生成查询"}
        queries_input = parsed_queries.queries
        query_results = await search_engine.abatch(
            queries_input, config, return_exceptions=True
        )
        successful_results = [
            res for res in query_results if not isinstance(res, Exception)
        ]
        all_query_results = {
            res["url"]: res["content"]
            for results in successful_results
            for res in results
        }
        # We could be more precise about handling max token length if we wanted to here
        dumped = json.dumps(all_query_results)[: self.max_str_len]
        ai_message: AIMessage = queries["raw"]
        tool_call = queries["raw"].additional_kwargs["tool_calls"][0]
        tool_id = tool_call["id"]
        tool_message = ToolMessage(tool_call_id=tool_id, content=dumped)
        swapped_state["messages"].extend([ai_message, tool_message])
        # Only update the shared state with the final answer to avoid
        # polluting the dialogue history with intermediate messages

        llm = self.runnable
        # gen_answer_chain = gen_answer_prompt | llm.with_structured_output(
        #     AnswerWithCitations, include_raw=True
        # ).with_config(run_name="GenerateAnswer").with_retry(stop_after_attempt=5)
        # generated = await gen_answer_chain.ainvoke(swapped_state)
        generated = await self.gen_answer(swapped_state, llm)
        cited_urls = set(generated["parsed"].cited_urls)
        if cited_urls:
            # Save the retrieved information to a the shared state for future reference
            cited_references = {
                k: v for k, v in all_query_results.items() if k in cited_urls
            }
        else:
            logger.warning("没有生成引用")
            cited_references = {}
        formatted_message = AIMessage(
            name=self.name, content=generated["parsed"].as_str
        )

        logger.info(f"gen_answer node 结束: {self.name}")
        return {"messages": [formatted_message], "references": cited_references}

    async def gen_answer(self, swapped_state, llm: Runnable):
        gen_answer_prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    """You are an expert who can use information effectively. You are chatting with a Wikipedia writer who wants\
        to write a Wikipedia page on the topic you know. You have gathered the related information and will now use the information to form a response.

        Make your response as informative as possible and make sure every sentence is supported by the gathered information.
        Each response must be backed up by a citation from a reliable source, formatted as a footnote, reproducing the URLS after your response.""",
                ),
                MessagesPlaceholder(variable_name="messages", optional=True),
            ]
        )

        # parser = PydanticOutputParser(pydantic_object=AnswerWithCitations)
        # retry_parser = RetryOutputParser.from_llm(parser=parser, llm=llm)

        gen_answer_chain = gen_answer_prompt | llm.with_structured_output(
            AnswerWithCitations, include_raw=True
        ).with_config(run_name="GenerateAnswer").with_retry(stop_after_attempt=5)

        max_retries = 5
        for attempt in range(max_retries):
            try:
                generated = await gen_answer_chain.ainvoke(swapped_state)
                if generated is None or not isinstance(
                    generated.get("parsed"), AnswerWithCitations
                ):
                    raise ValueError("Invalid response format")
                return generated
            except Exception as e:
                logger.warning(f"Attempt {attempt + 1} failed: {str(e)}")
                if attempt == max_retries - 1:
                    logger.error("Max retries reached. Returning error.")
                    return {
                        "error": "Failed to generate a valid response after multiple attempts"
                    }
                await asyncio.sleep(20)

    async def gen_queries(self, swapped_state, llm: Runnable):
        max_retries = 5
        for attempt in range(max_retries):
            try:
                gen_queries_prompt = ChatPromptTemplate.from_messages(
                    [
                        (
                            "system",
                            "You are a helpful research assistant. Query the search engine to answer the user's questions.",
                        ),
                        MessagesPlaceholder(variable_name="messages", optional=True),
                    ]
                )
                gen_queries_chain = (
                    gen_queries_prompt
                    | self.runnable.with_structured_output(
                        Queries, include_raw=True
                    ).with_retry(stop_after_attempt=5)
                )

                queries = await gen_queries_chain.ainvoke(swapped_state)

                if queries is None or not isinstance(queries.get("parsed"), Queries):
                    raise ValueError("生成搜索查询失败")
                return queries
            except Exception as e:
                logger.warning(f"Attempt {attempt + 1} failed: {str(e)}")
                if attempt == max_retries - 1:
                    logger.error("Max retries reached. Raising the last exception.")
                    raise
                await asyncio.sleep(20)



from dataclasses import dataclass
from typing import Any, Dict, Optional

from autogen_agentchat.agents import AssistantAgent
from autogen_core import (
    AgentId,
    DefaultTopicId,
    MessageContext,
    RoutedAgent,
    SingleThreadedAgentRuntime,
    default_subscription,
    message_handler,
)
from autogen_core.model_context import BufferedChatCompletionContext
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    SystemMessage,
    UserMessage,
)

from .model_client import get_oai_Model
from .termination_handler import termination_handler


class LoggingAssistantAgent(AssistantAgent):
    """扩展AssistantAgent以添加日志功能"""

    async def _process_received_message(self, message: Dict[str, Any]) -> Optional[str]:
        return await super()._process_received_message(message)

    async def run(self, *args, **kwargs):
        try:
            result = await super().run(*args, **kwargs)
            return result
        except Exception:
            raise


@dataclass
class Message:
    content: str


@default_subscription
class Assistant(RoutedAgent):
    def __init__(self, name: str, model_client: ChatCompletionClient) -> None:
        super().__init__("An assistant agent.")
        self._model_client = model_client
        self.name = name
        self.count = 0
        self._system_messages = [
            SystemMessage(
                content=f"Your name is {name} and you are a part of a duo of comedians."
                "You laugh when you find the joke funny, else reply 'I need to go now'.",
            )
        ]
        self._model_context = BufferedChatCompletionContext(buffer_size=5)

    @message_handler
    async def handle_message(self, message: Message, ctx: MessageContext) -> None:
        self.count += 1
        await self._model_context.add_message(
            UserMessage(content=message.content, source="user")
        )
        result = await self._model_client.create(
            self._system_messages + await self._model_context.get_messages()
        )

        print(f"\n{self.name}: {message.content}")

        if "I need to go".lower() in message.content.lower() or self.count > 2:
            return

        await self._model_context.add_message(
            AssistantMessage(content=result.content, source="assistant")
        )  # type: ignore
        await self.publish_message(Message(content=result.content), DefaultTopicId())  # type: ignore


async def hello_ag_run() -> None:
    agent = LoggingAssistantAgent(
        "blog_writer",
        model_client=get_oai_Model(),
        system_message="""你是一位极具洞察力的科技博主，擅长:
        1. 发现和分享AI领域最新、最有趣的应用和趋势
        2. 用生动的案例和数据支撑观点
        3. 提供独特的见解和实用的建议
        4. 善于用故事化的写作方式吸引读者
        请确保内容既专业又有趣，让读者感受到实用价值。""",
    )

    blog_prompt = """请写一篇引人入胜的博客文章，要求：
    1. 主题：选择以下任一个具体角度切入AI在日常生活中的应用：
       - "我用AI重新设计了我的早晨routine，发生了这些改变..."
       - "为什么说2024年是普通人入门AI的最佳时机"
       - "我用AI帮我做副业赚钱的一个月：经验和教训"
    2. 写作要求：
       - 开头要有吸引人的hook，比如一个有趣的故事或令人意外的数据
       - 分享真实可行的经验和洞察，而不是泛泛而谈
       - 每个观点都要配合具体的例子或数据
       - 提供实用的行动建议，读者看完就能上手尝试
       - 结尾要有清晰的call-to-action
    3. 结构：
       - 引言：吸引人的开场 + 文章价值预告
       - 正文：3-4个核心观点，每个配有实例
       - 总结：核心观点提炼 + 行动建议
    4. 字数：1500字左右，确保内容充实
    5. 风格：轻松但专业，像朋友分享经验一样自然
    """

    try:
        # 直接调用 agent
        result = await agent.run(task=blog_prompt)
        print(result)
        # llm_usage 有额外的跟踪信息
    except Exception:
        # logger.error("生成文章时发生错误", exc_info=True)
        raise

    # 通过 runtime 调用agent
    runtime = SingleThreadedAgentRuntime(intervention_handlers=[termination_handler])
    cathy = await Assistant.register(
        runtime,
        "cathy",
        lambda: Assistant(name="Cathy", model_client=get_oai_Model()),
    )

    joe = await Assistant.register(
        runtime,
        "joe",
        lambda: Assistant(name="Joe", model_client=get_oai_Model()),
    )

    runtime.start()
    await runtime.send_message(
        Message("Joe, tell me a joke."),
        recipient=AgentId(joe, "default"),
        sender=AgentId(cathy, "default"),
    )
    await runtime.stop_when_idle()



from typing import Literal

import structlog
from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import StreamingResponse
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.completion_create_params import CompletionCreateParamsBase
from opentelemetry import trace
from pydantic import BaseModel

# from mtmai.agents.chat_profiles.profiles import get_chat_agent
# from mtmai.agents.ctx import init_mtmai_http_context, mtmai_context
# from mtmai.agents.graphs.hello_graph import HelloGraph
# from mtmai.agents.opencanvas.opencanvas_graph import OpenCanvasGraph
# from mtmai.agents.sitegen_graph.sitegen_graph import SiteGenGraph
# from mtmai.assistants.assistants import get_assistant_agent
from mtmai.deps import OptionalUserDep
from mtmai.llm.llm import (
    call_chat_completions,
    chat_completions_stream_generator,
    default_llm_model,
)

router = APIRouter()
LOG = structlog.get_logger()

tracer = trace.get_tracer_provider().get_tracer(__name__)


API_KEY_NAME = "Authorization"


class MtmaiCompletionRequest(CompletionCreateParamsBase):
    class Config:
        arbitrary_types_allowed = True

    chatId: bool | None = False
    threadId: str | None = None
    isChat: bool | None = False
    messages: list[ChatCompletionMessageParam] | None = []
    model: str | None = None
    stream: bool | None = False


default_chat_profile = "taskrunner"


async def context_sse_stream():
    while True:
        message = await mtmai_context.get_next_event()
        if message:
            yield f"data: {message}\n\n"  # Format the message for SSE


@router.post("/chat/completions")
async def chat_completions(
    request: Request,
    user: OptionalUserDep,
):
    """以兼容 openai completions 协议的方式 反代其他 ai 提供商的 completions api"""
    request_data = await request.json()

    prompt = request_data.get("prompt")
    if prompt:
        request_data["messages"] = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        request_data.pop("prompt")

    is_stream = request_data.get("stream")

    model = request_data.get("model") or default_llm_model
    is_chat = request.headers.get("X-Chat") == "1" or request_data.get("isChat")

    init_mtmai_http_context(
        thread_id=request_data.get("threadId"),
        user_id=str(user.id) if user else None,
    )
    is_pull = request_data.get("isPull")

    if is_pull:
        await mtmai_context.init_mq()
        return StreamingResponse(
            mtmai_context.mq.pull_messages(),
            media_type="text/event-stream",
        )

    request_data["model"] = model
    messages = request_data.get("messages")
    if not is_chat:
        # open ai 兼容的模型调用
        if not messages:
            return HTTPException(status_code=503, detail="No messages provided")
        if is_stream:
            return StreamingResponse(
                chat_completions_stream_generator(request_data),
                media_type="text/event-stream",
            )

        else:
            return await call_chat_completions(request_data)

    params = request_data.get("params")
    chat_profile = request_data.get("chatProfile", default_chat_profile)
    if chat_profile == "hello_graph":
        hello_graph = HelloGraph()
        return StreamingResponse(
            hello_graph.run_graph(
                messages=messages,  # user_id=str(user.id), params=params
            ),
            media_type="text/event-stream",
        )
    if chat_profile == "opencanvas":  # 未完成
        hello_graph = OpenCanvasGraph()
        return StreamingResponse(
            hello_graph.run_graph(
                messages=messages,  # user_id=str(user.id), params=params
                user_id=str(user.id) if user else None,
                params=params,
            ),
            media_type="text/event-stream",
        )
    if chat_profile == "sitegen":
        graph = SiteGenGraph()
        return StreamingResponse(
            graph.run_graph(
                messages=messages,  # user_id=str(user.id), params=params
                user_id=str(user.id) if user else None,
                params=params,
            ),
            media_type="text/event-stream",
        )

    if chat_profile == "taskrunner":
        graph_app = await get_chat_agent("taskrunner")

        graph_app.onRequest(messages=messages, inputs={})
        # 这里不用返回数据，因为现在使用 redis 消息队列。内部会自动向队列发送消息。
        # return StreamingResponse(
        #     context_sse_stream,
        #     media_type="text/event-stream",
        # )

    else:
        assistant_agent = await get_assistant_agent(chat_profile)
        return StreamingResponse(
            assistant_agent.stream_messages(
                messages=messages, user_id=str(user.id), params=params
            ),
            media_type="text/event-stream",
            headers={
                # "x-vercel-ai-data-stream": "v1",
                "Content-Type": "text/plain; charset=utf-8",
                "Vary": "RSC, Next-Router-State-Tree, Next-Router-Prefetch",
                "Cache-Control": "no-store, no-cache, must-revalidate, proxy-revalidate",
                "Pragma": "no-cache",
                "Expires": "0",
                "Surrogate-control": "no-store",
            },
        )


class ChatEventRequest(BaseModel):
    thread_id: str


class ChatEventItem(BaseModel):
    role: Literal["user", "assistant"]
    content: str


class ChatEventResponse(BaseModel):
    items: list[ChatEventItem]


# import os
# from shutil import rmtree

# import torch
# from fastapi import APIRouter, Request
# from langchain_huggingface.embeddings import HuggingFaceEmbeddings
# from optimum.onnxruntime import ORTModelForFeatureExtraction
# from pydantic import BaseModel, Field
# from starlette.concurrency import run_in_threadpool
# from transformers import AutoTokenizer

# from mtmai.mtlibs.logging import get_logger

# router = APIRouter()

# logger = logging.getLogger()
# tokenizer = None
# model_current = None
# device = None
# provider = None


# class CreateEmbeddingRequest(BaseModel):
#     input: str | list[str] = Field(description="The input to embed.")
#     model: str | None
#     # input_type: Optional[str]  # 可能没用
#     encoding_format: str | None


# class Embedding(BaseModel):
#     embedding: list[float] = Field(max_length=2048)


# class CreateEmbeddingResponse(BaseModel):
#     data: list[Embedding]


# def save_model(model_name, _device, _reload=False):
#     global device, provider
#     device = _device
#     provider = "CUDAExecutionProvider" if device == "cuda" else "CPUExecutionProvider"
#     optimize = "O4" if device == "cuda" else "O3"
#     ort_model_output = f".vol/model_auto_opt_{optimize}"
#     model_exists = os.path.exists(ort_model_output) and os.path.exists(
#         ort_model_output + "/config.json"
#     )
#     if (model_exists and _reload) or not model_exists:
#         if model_exists:
#             rmtree(ort_model_output)
#         print(f"Loading {model_name} model and exporting it to ONNX..")
#         _model = ORTModelForFeatureExtraction.from_pretrained(
#             model_name, provider=provider, export=True, cache_dir="/tmp"
#         )
#         _model.save_pretrained(ort_model_output)
#     elif model_exists and not _reload:
#         print(f"Model {model_name} already exists! use RELOAD=True to reload")
#     return "./" + ort_model_output


# def load_model(model_name, _device, _reload):
#     global tokenizer, model_current
#     ort_model_output = save_model(model_name, _device, _reload=_reload)
#     print(f"Loading {model_name} model from {ort_model_output}..")
#     tokenizer = AutoTokenizer.from_pretrained(ort_model_output)
#     model_current = ORTModelForFeatureExtraction.from_pretrained(
#         ort_model_output, provider=provider
#     )


# def _create_embedding(input: str | list[str], model: str, encoding_format: str | None):
#     global tokenizer
#     if not isinstance(input, list):
#         input = [input]

#     if not model_current:
#         print("加载模型", model)
#         load_model(model, "cpu", False)
#     encoded_input = tokenizer(
#         input, padding=True, truncation=True, return_tensors="pt"
#     ).to(device)

#     # 使用 torch 属于比较底层的操作。
#     with torch.no_grad():
#         sentence_embeddings = model_current(**encoded_input)[0][:, 0]
#     sentence_embeddings = (
#         torch.nn.functional.normalize(sentence_embeddings, p=2, dim=1).cpu().numpy()
#     )
#     data = [Embedding(embedding=embedding) for embedding in sentence_embeddings]
#     return CreateEmbeddingResponse(data=data)
#     # 也可以使用 高级库，例如：
#     # embModel = getEmbdingModel()
#     # texts = [input]
#     # nums=embModel.embed_documents(texts)


# @router.post(
#     "/v1/embeddings",
#     #  response_model=CreateEmbeddingResponse,
# )
# async def create_embedding(
#     # request: CreateEmbeddingRequest,
#     request: Request,
# ):
#     """
#     供前端获取计算 embdings （目前使用本地模型）
#     api 接口可以参考： https://platform.openai.com/docs/guides/embeddings/what-are-embeddings
#     """
#     # print("v1 embeddings 调用:", request.model)
#     aa = await request.json()
#     print(aa)
#     return await run_in_threadpool(_create_embedding, **aa)


# embdingModel = None


# def getEmbdingModel():
#     """
#     文档： 嵌入模型排行榜 https://huggingface.co/spaces/mteb/leaderboard
#     """
#     global embdingModel
#     if embdingModel is None:
#         # model_name = "mixedbread-ai/mxbai-embed-large-v1"
#         # model_name = "bert-base-uncased" #维度768，# 体积约0.5G
#         # model_name = "microsoft/Phi-3-mini-4k-instruct" # 维度 3072 模型文件约8G，性能上的消耗比较大。模型加载时长约20s
#         # model_name = "iampanda/zpoint_large_embedding_zh" # 中文 维度：1792
#         # model_name = "voyageai/voyage-lite-02-instruct" # 1536
#         model_name = "infgrad/stella-large-zh-v2"  # 1024
#         hf_embeddings = HuggingFaceEmbeddings(
#             model_name=model_name,
#             # trust_remote_code=True
#         )
#         embdingModel = hf_embeddings
#     return embdingModel



import asyncio
import os
import threading
import uuid
from datetime import datetime
from typing import Awaitable, Callable

import structlog
import uvicorn
from dotenv import load_dotenv
from fastapi import APIRouter, Request
from fastapi.concurrency import asynccontextmanager
from fastapi.responses import JSONResponse
from fastapi.routing import APIRoute
from mtmai import analytics
from mtmai.app_forge import setup_forge_app
from mtmai.core import coreutils
from mtmai.core.__version__ import version
from mtmai.core.config import settings
from mtmai.core.coreutils import is_in_dev, is_in_vercel
from mtmai.exceptions import SkyvernHTTPException
from mtmai.forge.sdk.db.exceptions import NotFoundError
from mtmai.forge.sdk.settings_manager import SettingsManager
from mtmai.middleware import AuthMiddleware
from mtmaisdk.utils.env import is_in_docker, is_in_huggingface, is_in_windows
from mtmlib import mtutils
from pydantic import ValidationError
from starlette.requests import HTTPConnection
from starlette.requests import Request as StarletteRequest
from starlette.templating import Jinja2Templates
from starlette_context.plugins.base import Plugin

# from mtmai.workflows.workers import deploy_mtmai_workers

LOG = structlog.stdlib.get_logger()


class ExecutionDatePlugin(Plugin):
    key = "execution_date"

    async def process_request(
        self, request: StarletteRequest | HTTPConnection
    ) -> datetime:
        return datetime.now()


def build_app():
    from fastapi import FastAPI

    LOG.info("build api app.")

    LOG.info("api auth")
    api_router = APIRouter()
    from mtmai.api import auth

    api_router.include_router(auth.router, tags=["auth"])
    LOG.info("api users")

    from mtmai.api import users

    api_router.include_router(users.router, prefix="/users", tags=["users"])

    LOG.info("api blog")
    from mtmai.api import blog

    api_router.include_router(blog.router, prefix="/posts", tags=["posts"])

    LOG.info("api image")
    from mtmai.api import image

    api_router.include_router(image.router, prefix="/image", tags=["image"])

    LOG.info("api train")
    from mtmai.api import train

    api_router.include_router(train.router, prefix="/train", tags=["train"])

    LOG.info("api metrics")
    from mtmai.api import metrics

    api_router.include_router(metrics.router, prefix="/metrics", tags=["metrics"])

    # LOG.info("api agent")
    # from mtmai.api import agent

    # api_router.include_router(agent.router, prefix="/agent", tags=["agent"])

    LOG.info("api agent")
    from mtmai.api import agent

    api_router.include_router(agent.router, prefix="/agent", tags=["agent"])

    # LOG.info("api form")
    # from mtmai.api import form

    # api_router.include_router(form.router, prefix="/form", tags=["form"])

    LOG.info("api site")
    from mtmai.api import site

    api_router.include_router(site.router, prefix="/site", tags=["site"])

    LOG.info("api webpage")
    from mtmai.api import webpage

    api_router.include_router(webpage.router, prefix="/webpage", tags=["webpage"])

    LOG.info("api tasks")
    from mtmai.api import tasks

    api_router.include_router(tasks.router, prefix="/tasks", tags=["tasks"])

    LOG.info("api openai")
    from mtmai.api import openai

    api_router.include_router(openai.router, tags=["openai"])

    LOG.info("api listview")
    from mtmai.api import listview

    api_router.include_router(listview.router, prefix="/listview", tags=["listview"])

    LOG.info("api workbench")
    from mtmai.api import workbench

    api_router.include_router(workbench.router, prefix="/workbench", tags=["workbench"])

    LOG.info("api logs")
    from mtmai.api import logs

    api_router.include_router(logs.router, prefix="/logs", tags=["logs"])

    LOG.info("api thread")
    from mtmai.api import thread

    api_router.include_router(thread.router, prefix="/thread", tags=["thread"])

    LOG.info("api artifact")
    from mtmai.api import artifact

    api_router.include_router(artifact.router, prefix="/artifact", tags=["artifact"])

    LOG.info("api config")
    from mtmai.api import config

    api_router.include_router(config.router, prefix="/config", tags=["config"])

    # LOG.info("api chat2")
    # from mtmai.api import chat2

    # api_router.include_router(chat2.router, prefix="/chat2", tags=["chat2"])

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        # deploy_mtmai_workers()
        yield
        # await taskWroker.stop()

    def custom_generate_unique_id(route: APIRoute) -> str:
        if len(route.tags) > 0:
            return f"{route.tags[0]}-{route.name}"
        return f"{route.name}"

    openapi_tags = [
        {
            "name": "admin",
            "description": "管理专用 ",
        },
        {
            "name": "train",
            "description": "模型训练及数据集",
        },
        {
            "name": "mtmcrawler",
            "description": "爬虫数据采集 ",
        },
        {
            "name": "openai",
            "description": "提供兼容 OPEN AI 协议 , 外置工作流 例如 langflow 可以通过此endpoint调用内部的工作流和模型",
        },
    ]

    app = FastAPI(
        # docs_url=None,
        # redoc_url=None,
        title=settings.PROJECT_NAME,
        description="mtmai description(group)",
        version=version,
        lifespan=lifespan,
        generate_unique_id_function=custom_generate_unique_id,
        openapi_url=f"{settings.API_V1_STR}/openapi.json",
        swagger_ui_parameters={
            "syntaxHighlight": True,
            "syntaxHighlight.theme": "obsidian",
        },
        openapi_tags=openapi_tags,
    )
    templates = Jinja2Templates(directory="templates")

    if is_in_dev():
        from mtmai.api import admin

        api_router.include_router(
            admin.router,
            prefix="/admin",
            tags=["admin"],
        )
        # from mtmai.api import demos

        # api_router.include_router(
        #     demos.router, prefix="/demos/demos", tags=["demos_demos"]
        # )

    # app.openapi_schema = {
    #     "components": {
    #         "schemas": {
    #             "MessagePayload": MessagePayload.model_json_schema(),
    #             "AudioChunkPayload": AudioChunkPayload.model_json_schema(),
    #         }
    #     }
    # }

    @app.exception_handler(Exception)
    async def generic_exception_handler(request: Request, exc: Exception):  # noqa: ARG001
        return JSONResponse(status_code=500, content={"detail": str(exc)})

    def setup_main_routes():
        from mtmai.api import home

        app.include_router(home.router)
        app.include_router(api_router, prefix=settings.API_V1_STR)

    setup_main_routes()

    if settings.OTEL_ENABLED:
        from mtmai.mtlibs import otel

        otel.setup_otel(app)

    if settings.BACKEND_CORS_ORIGINS:
        from fastapi.middleware.cors import CORSMiddleware

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"]
            if settings.BACKEND_CORS_ORIGINS == "*"
            else [str(origin).strip("/") for origin in settings.BACKEND_CORS_ORIGINS],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*", "x-chainlit-client-type"],
        )
        app.add_middleware(AuthMiddleware)

    setup_forge_app(app)

    from starlette_context.middleware import RawContextMiddleware

    app.add_middleware(
        RawContextMiddleware,
        plugins=(
            # TODO (suchintan): We should set these up
            ExecutionDatePlugin(),
            # RequestIdPlugin(),
            # UserAgentPlugin(),
        ),
    )

    from fastapi import FastAPI, Response, status

    @app.exception_handler(NotFoundError)
    async def handle_not_found_error(request: Request, exc: NotFoundError) -> Response:
        return Response(status_code=status.HTTP_404_NOT_FOUND)

    @app.exception_handler(SkyvernHTTPException)
    async def handle_skyvern_http_exception(
        request: Request, exc: SkyvernHTTPException
    ) -> JSONResponse:
        return JSONResponse(
            status_code=exc.status_code, content={"detail": exc.message}
        )

    @app.exception_handler(ValidationError)
    async def handle_pydantic_validation_error(
        request: Request, exc: ValidationError
    ) -> JSONResponse:
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content={"detail": str(exc)},
        )

    @app.exception_handler(Exception)
    async def unexpected_exception(request: Request, exc: Exception) -> JSONResponse:
        LOG.exception("Unexpected error in agent server.", exc_info=exc)
        return JSONResponse(
            status_code=500, content={"error": f"Unexpected error: {exc}"}
        )

    from mtmai.forge.sdk.core import skyvern_context
    from mtmai.forge.sdk.core.skyvern_context import SkyvernContext

    @app.middleware("http")
    async def request_middleware(
        request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        curr_ctx = skyvern_context.current()
        if not curr_ctx:
            request_id = str(uuid.uuid4())
            skyvern_context.set(SkyvernContext(request_id=request_id))
        elif not curr_ctx.request_id:
            curr_ctx.request_id = str(uuid.uuid4())

        try:
            return await call_next(request)
        finally:
            skyvern_context.reset()

    if SettingsManager.get_settings().ADDITIONAL_MODULES:
        for module in SettingsManager.get_settings().ADDITIONAL_MODULES:
            LOG.info("Loading additional module to set up api app", module=module)
            __import__(module)
        LOG.info(
            "Additional modules loaded to set up api app",
            modules=SettingsManager.get_settings().ADDITIONAL_MODULES,
        )

    from mtmai.forge import app as forge_app

    if forge_app.setup_api_app:
        forge_app.setup_api_app(app)

    return app


async def serve():
    try:
        analytics.capture("skyvern-oss-run-server")
        load_dotenv()

        app = build_app()
        # threading.Thread(target=start_deamon_serve).start()

        config = uvicorn.Config(
            app,
            host=settings.SERVE_IP,
            port=settings.PORT,
            log_level="info",
            # reload=not settings.is_production,
            # !!! bug 当 使用了prefect 后，使用了这个指令： @flow，程序会被卡在： .venv/lib/python3.12/site-packages/uvicorn/config.py
            # !!!365行：logging.config.dictConfig(self.log_config)
            # !!! 原因未知，但是 log_config=None 后，问题消失
            # log_config=None,
            # reload=reload,
        )

        host = (
            "127.0.0.1"
            if settings.SERVE_IP == "0.0.0.0"
            else settings.server_host.split("://")[-1]
        )
        server_url = f"{settings.server_host.split('://')[0]}://{host}:{settings.PORT}"

        LOG.info(
            "server config.", host="0.0.0.0", port=settings.PORT, server_url=server_url
        )
        server = uvicorn.Server(config)

        try:
            LOG.info(
                "server starting",
                host="0.0.0.0",
                port=settings.PORT,
                server_url=server_url,
            )
            await server.serve()
        except Exception as e:
            LOG.error("Error in uvicorn server:", exc_info=e)
            raise

    except Exception as e:
        LOG.error("server error: %s", e)


def start_deamon_serve():
    """
    启动后台独立服务
    根据具体环境自动启动
    """
    LOG.info("start_deamon_serve")
    if is_in_dev():
        from mtmai.flows.deployments import start_prefect_deployment

        start_prefect_deployment(asThreading=True)

    if (
        not settings.is_in_vercel
        and not settings.is_in_gitpod
        and settings.CF_TUNNEL_TOKEN
        and not is_in_huggingface()
        and not is_in_windows()
    ):
        from mtmlib import tunnel

        threading.Thread(target=lambda: asyncio.run(tunnel.start_cloudflared())).start()

        if not is_in_vercel() and not settings.is_in_gitpod:
            from mtmai.mtlibs.server.searxng import run_searxng_server

            threading.Thread(target=run_searxng_server).start()
        if (
            not settings.is_in_vercel
            and not settings.is_in_gitpod
            and not is_in_windows()
        ):

            def start_front_app():
                mtmai_url = coreutils.backend_url_base()
                if not mtutils.command_exists("mtmaiweb"):
                    LOG.warning("⚠️ mtmaiweb 命令未安装,跳过前端的启动")
                    return
                mtutils.bash(
                    f"PORT={settings.FRONT_PORT} MTMAI_API_BASE={mtmai_url} mtmaiweb serve"
                )

            threading.Thread(target=start_front_app).start()

            def start_prefect_server():
                LOG.info("启动 prefect server")

                sqlite_db_path = "/app/storage/prefect.db"
                sql_connect_str = f"sqlite+aiosqlite:///{sqlite_db_path}"
                mtutils.bash(
                    f"PREFECT_UI_STATIC_DIRECTORY=/app/storage PREFECT_API_DATABASE_CONNECTION_URL={sql_connect_str} prefect server start"
                )

            threading.Thread(target=start_prefect_server).start()

        if not is_in_vercel() and not settings.is_in_gitpod and not is_in_windows():
            from mtmai.mtlibs.server.kasmvnc import run_kasmvnc

            threading.Thread(target=run_kasmvnc).start()

        if is_in_docker():
            from mtmai.mtlibs.server.easyspider import run_easy_spider_server

            threading.Thread(target=run_easy_spider_server).start()

    LOG.info("start deamon finished")



import json
from typing import Any, Optional, Type, TypeVar

from browser_use import Agent, AgentHistoryList
from browser_use.agent.views import (ActionResult, AgentError, AgentHistory,
                                     AgentHistoryList, AgentOutput,
                                     AgentStepInfo)
from browser_use.browser.browser import Browser, BrowserConfig
from browser_use.utils import time_execution_async
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage, SystemMessage
from langchain_openai import ChatOpenAI
from matplotlib import hatch
from mtmaisdk.clients.rest.models import BrowserParams
from mtmaisdk.context.context import Context

from mtmai.agents.ctx import get_mtmai_context, init_mtmai_context
from mtmai.mtlibs.httpx_transport import LoggingTransport
from mtmai.worker import wfapp


class BrowserAgent(Agent):
    pass
    # @time_execution_async('--get_next_action')
    # async def get_next_action(self, input_messages: list[BaseMessage]) -> AgentOutput:
    #     """Get next action from LLM based on current state"""

    #     # structured_llm = self.llm.with_structured_output(self.AgentOutput, include_raw=True)
    #     # response: dict[str, Any] = await structured_llm.ainvoke(input_messages)  # type: ignore  # noqa: F821
        
    #     content = AgentOutput.model_json_schema()
    #     systemToolPrompt = SystemMessage(content=f"VERY IMPORTANT: YOU MUST FOLLOW THE STRUCTURE OF THE JSON SCHEMA: \n{content}")
    #     input_messages.append(systemToolPrompt)
    #     response: AIMessage = await self.llm.ainvoke(input_messages)  # type: ignore
        
    #     json_data = json.loads(response.content)
    #     parsed = AgentOutput.model_validate(json_data)
    #     # parsed: AgentOutput = response['parsed']
    #     if parsed is None:
    #         raise ValueError('Could not parse response.')

    #     # cut the number of actions to max_actions_per_step
    #     parsed.action = parsed.action[: self.max_actions_per_step]
    #     self._log_response(parsed)
    #     self.n_steps += 1

    #     return parsed


import torch
from transformers import DistilBertTokenizer, DistilBertForSequenceClassification
from transformers import pipeline

# 固定随机数，可以确保相同环境下运算的结果相同。
seed_value = 42
torch.manual_seed(seed_value)

# 使用 GPU 如果可用
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# 加载预训练的 DistilBERT 分类模型
model_name = "distilbert-base-uncased"

pipeModel1 = pipeline(model="facebook/bart-large-mnli")

# 给定文章，改写文章，输出的结果将是最终文章。
def art_rewrite(title, content, candidate_labels):
    # test_input=""I have a problem with my iphone that needs to be resolved asap!!","
    classifyResult=pipeModel1(input_text, candidate_labels=candidate_labels,)
    # 挑选最匹配的一个
    max_score_index = classifyResult["scores"].index(max(classifyResult["scores"]))
    max_label = classifyResult["labels"][max_score_index]
    return max_label



from enum import StrEnum
from pathlib import Path

# This is the attribute name used to tag interactable elements
SKYVERN_ID_ATTR: str = "unique_id"
SKYVERN_DIR = Path(__file__).parent
REPO_ROOT_DIR = SKYVERN_DIR.parent

INPUT_TEXT_TIMEOUT = 120000  # 2 minutes
PAGE_CONTENT_TIMEOUT = 300  # 5 mins
BROWSER_CLOSE_TIMEOUT = 180  # 3 minute

# reserved fields for navigation payload
SPECIAL_FIELD_VERIFICATION_CODE = "verification_code"


class ScrapeType(StrEnum):
    NORMAL = "normal"
    STOPLOADING = "stoploading"
    RELOAD = "reload"


SCRAPE_TYPE_ORDER = [ScrapeType.NORMAL, ScrapeType.NORMAL, ScrapeType.RELOAD]



"""
工作流： 文章生成
"""

import asyncio
from datetime import timedelta
from textwrap import dedent

from crewai import Agent, Crew, Process, Task
from prefect import get_run_logger, task
from prefect.tasks import task_input_hash
from pydantic import BaseModel

from mtmai.agents.ctx import mtmai_context
from mtmai.agents.tools.tools import get_tools
from mtmai.models.book_gen import (
    BookOutline,
    Chapter,
    WriteOutlineRequest,
    WriteSingleChapterRequest,
)
from mtmai.mtlibs.aiutils import get_json_format_instructions


@task(
    cache_key_fn=task_input_hash,
    cache_expiration=timedelta(seconds=60),
    retries=3,
)
async def article_gen_outline(*, req: WriteOutlineRequest) -> BookOutline:
    """生成文章大纲"""
    llm = await mtmai_context.get_crawai_llm()
    logger = get_run_logger()

    researcher_agent = Agent(
        role="Research Agent",
        goal=dedent("""Gather comprehensive information about {topic} that will be used to create an organized and well-structured book outline.
Here is some additional information about the author's desired goal for the book:\n\n {goal}"""),
        backstory=dedent("""You're a seasoned researcher, known for gathering the best sources and understanding the key elements of any topic.
You aim to collect all relevant information so the book outline can be accurate and informative."""),
        tools=get_tools("search_engine"),
        llm=llm,
        verbose=True,
    )
    outliner_agent = Agent(
        role="Book Outlining Agent",
        goal=dedent("""Based on the research, generate a book outline about the following topic: {topic}
The generated outline should include all chapters in sequential order and provide a title and description for each chapter.
Here is some additional information about the author's desired goal for the book:\n\n {goal}"""),
        backstory=dedent("""You are a skilled organizer, great at turning scattered information into a structured format.
Your goal is to create clear, concise chapter outlines with all key topics and subtopics covered."""),
        llm=llm,
        verbose=True,
    )

    research_topic_task = Task(
        description=dedent("""Research the provided topic of {topic} to gather the most important information that will
be useful in creating a book outline. Ensure you focus on high-quality, reliable sources.

Here is some additional information about the author's desired goal for the book:\n\n {goal}
        """),
        expected_output="A set of key points and important information about {topic} that will be used to create the outline.",
        agent=researcher_agent,
    )
    generate_outline_task = Task(
        description=dedent("""Create a book outline with chapters in sequential order based on the research findings.
Ensure that each chapter has a title and a brief description that highlights the topics and subtopics to be covered.
It's important to note that each chapter is only going to be 3,000 words or less.
Also, make sure that you do not duplicate any chapters or topics in the outline.

Here is some additional information about the author's desired goal for the book:\n\n {goal}"""),
        expected_output="An outline of chapters, with titles and descriptions of what each chapter will contain. Maximum of 3 chapters.  \n\n {format_instructions}",
        output_pydantic=BookOutline,
        agent=outliner_agent,
    )
    crew = Crew(
        agents=[researcher_agent, outliner_agent],
        tasks=[research_topic_task, generate_outline_task],
        process=Process.sequential,
        verbose=True,
    )

    inputs = req
    if isinstance(req, BaseModel):
        inputs = req.model_dump()
    inputs["format_instructions"] = get_json_format_instructions(BookOutline)
    output = await crew.kickoff_async(inputs=inputs)

    logger.info(f"format_instructions: {get_json_format_instructions(BookOutline)}")
    if not output:
        raise ValueError(f"大纲生成失败: {req}")
    if not output.pydantic:
        raise ValueError(
            f"大纲生成失败: 原因是 output.pydantic 没有正确的输出格式,原始内容: {output.raw}"
        )
    return output.pydantic


@task(
    cache_key_fn=task_input_hash,
    cache_expiration=timedelta(days=1),
    retries=3,
)
async def write_book_chapter_crew(*, req: WriteSingleChapterRequest):
    """生成文章一个章节"""
    logger = get_run_logger()

    llm = await mtmai_context.get_crawai_llm()

    researcher_agent = Agent(
        role="Research Agent",
        goal=dedent("""Gather comprehensive information about {topic} and {chapter_title} that will be used to enhance the content of the chapter.
Here is some additional information about the author's desired goal for the book and the chapter:\n\n {goal}
Here is the outline description for the chapter:\n\n {chapter_description}"""),
        backstory=dedent("""You are an experienced researcher skilled in finding the most relevant and up-to-date information on any given topic.
Your job is to provide insightful data that supports and enriches the writing process for the chapter."""),
        # tools=[],
        tools=get_tools("search_engine"),
        llm=llm,
    )

    writer_agent = Agent(
        role="Chapter Writer",
        goal=dedent("""Write a well-structured chapter for the book based on the provided chapter title, goal, and outline.
The chapter should be written in markdown format and contain around 3,000 words."""),
        backstory=dedent("""You are an exceptional writer, known for producing engaging, well-researched, and informative content.
You excel at transforming complex ideas into readable and well-organized chapters."""),
        llm=llm,
    )

    research_chapter_task = Task(
        description=dedent("""Research the provided chapter topic, title, and outline to gather additional content that will be helpful in writing the chapter.
Ensure you focus on reliable, high-quality sources of information.

Here is some additional information about the author's desired goal for the book and the chapter:\n\n {goal}
Here is the outline description for the chapter:\n\n {chapter_description}

When researching, consider the following key points:
- you need to gather enough information to write a 3,000-word chapter
- The chapter you are researching needs to fit in well with the rest of the chapters in the book.

Here is the outline of the entire book:\n\n
{book_outlines}"""),
        expected_output="A set of additional insights and information that can be used in writing the chapter.",
        agent=researcher_agent,
    )

    write_chapter_task = Task(
        description=dedent("""Write a well-structured chapter based on the chapter title, goal, and outline description.
Each chapter should be written in markdown and should contain around 3,000 words.

Here is the topic for the book: {topic}
Here is the title of the chapter: {chapter_title}
Here is the outline description for the chapter:\n\n {chapter_description}

Important notes:
- The chapter you are writing needs to fit in well with the rest of the chapters in the book.

Here is the outline of the entire book:\n\n
{book_outlines}"""),
        agent=writer_agent,
        expected_output="A markdown-formatted chapter of around 3,000 words that covers the provided chapter title and outline description. \n\n {format_instructions}",
        output_pydantic=Chapter,
    )
    crew = Crew(
        agents=[researcher_agent, writer_agent],
        tasks=[research_chapter_task, write_chapter_task],
        process=Process.sequential,
        verbose=True,
    )
    inputs = req.model_dump()
    inputs["format_instructions"] = get_json_format_instructions(Chapter)
    output = await crew.kickoff_async(inputs=inputs)
    if not output:
        raise ValueError(f"章节生成失败: {req}")
    if not output.pydantic:
        raise ValueError(
            f"章节生成失败,原因是 output.pydantic 没有正确的输出格式,原始内容: \n========\n{output.raw}\n========\n"
        )
    return output.pydantic


@task(cache_key_fn=task_input_hash, cache_expiration=timedelta(days=1))
async def write_section(topic):
    logger = get_run_logger()
    logger.info(f"开始写内容:{topic}")
    await asyncio.sleep(5)
    logger.info(f"写内容完成:{topic}")

    return f"Content for {topic}"


class ArticleGenStateRequest(BaseModel):
    topic: str | None = None
    user_id: str | None = None
    site_id: str | None = None



from datetime import timedelta
from typing import Annotated
from urllib.parse import urlencode

import httpx
import structlog
from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import JSONResponse, RedirectResponse
from fastapi.security import OAuth2PasswordRequestForm

from mtmai.core import coreutils, security
from mtmai.core.config import settings
from mtmai.core.security import get_password_hash
from mtmai.crud import curd, curd_account

# from mtmai.deps import AsyncSessionDep, SessionDep, get_current_active_superuser
# from mtmai.forge.app import DATABASE
# from mtmai.forge.sdk.services.org_auth_token_service import create_org_api_token
from mtmai.models.models import Message, NewPassword, Token

from ..deps import AsyncSessionDep, SessionDep

# from mtmai.utils import (  # generate_password_reset_token,; send_email,
#     # generate_reset_password_email,
#     # verify_password_reset_token,
# )

# from .utils.github import get_github_user_data


router = APIRouter()
LOG = structlog.get_logger()


@router.post("/login/access-token")
async def login_access_token(
    session: AsyncSessionDep, form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
) -> Token:
    """
    OAuth2 compatible token login, get an access token for future requests
    """
    user = await curd.authenticate(
        session=session, email=form_data.username, password=form_data.password
    )
    if not user:
        raise HTTPException(status_code=400, detail="Incorrect email or password")
    if not user.is_active:
        raise HTTPException(status_code=400, detail="Inactive user")
    access_token_expires = timedelta(minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = Token(
        access_token=security.create_access_token(
            user.id, expires_delta=access_token_expires
        )
    )

    response = JSONResponse(content=access_token.model_dump())

    response.set_cookie(
        key=settings.COOKIE_ACCESS_TOKEN,
        value=access_token.access_token,
        httponly=True,
        secure=True,
        samesite="lax",
        max_age=settings.ACCESS_TOKEN_EXPIRE_MINUTES * 60,
    )

    return response


# @router.post("/password-recovery/{email}")
# async def recover_password(email: str, session: AsyncSessionDep) -> Message:
#     """
#     Password Recovery
#     """
#     user = await curd.get_user_by_email(session=session, email=email)

#     if not user:
#         raise HTTPException(
#             status_code=404,
#             detail="The user with this email does not exist in the system.",
#         )
#     password_reset_token = generate_password_reset_token(email=email)
#     email_data = generate_reset_password_email(
#         email_to=user.email, email=email, token=password_reset_token
#     )
#     send_email(
#         email_to=user.email,
#         subject=email_data.subject,
#         html_content=email_data.html_content,
#     )
#     return Message(message="Password recovery email sent")


@router.post("/reset-password/")
async def reset_password(session: AsyncSessionDep, body: NewPassword) -> Message:
    """
    Reset password
    """
    email = verify_password_reset_token(token=body.token)
    if not email:
        raise HTTPException(status_code=400, detail="Invalid token")
    user = await curd.get_user_by_email(session=session, email=email)
    if not user:
        raise HTTPException(
            status_code=404,
            detail="The user with this email does not exist in the system.",
        )
    if not user.is_active:
        raise HTTPException(status_code=400, detail="Inactive user")
    hashed_password = get_password_hash(password=body.new_password)
    user.hashed_password = hashed_password
    session.add(user)
    session.commit()
    return Message(message="Password updated successfully")


# @router.post(
#     "/password-recovery-html-content/{email}",
#     dependencies=[Depends(get_current_active_superuser)],
#     response_class=HTMLResponse,
# )
# async def recover_password_html_content(email: str, session: SessionDep) -> Any:
#     """
#     HTML Content for Password Recovery
#     """
#     user = await curd.get_user_by_email(session=session, email=email)

#     if not user:
#         raise HTTPException(
#             status_code=404,
#             detail="The user with this username does not exist in the system.",
#         )
#     password_reset_token = generate_password_reset_token(email=email)
#     email_data = generate_reset_password_email(
#         email_to=user.email, email=email, token=password_reset_token
#     )

#     return HTMLResponse(
#         content=email_data.html_content, headers={"subject:": email_data.subject}
#     )


@router.get("/auth/github/callback", include_in_schema=False)
async def github_oauth_callback(db: SessionDep, req: Request):
    # 打印请求的 query 参数
    print(f"Full URL: {req.url}")
    print(f"Query parameters: {req.query_params}")
    # 从请求中获取 GitHub OAuth 传回的 code 和 state 参数
    code = req.query_params.get("code")
    state = req.query_params.get("state", "")

    if not code:
        raise HTTPException(status_code=400, detail="Code not provided")

    client_id = settings.GITHUB_CLIENT_ID
    client_secret = settings.GITHUB_CLIENT_SECRET
    # 交换 code 获取 access token
    async with httpx.AsyncClient() as client:
        token_response = await client.post(
            "https://github.com/login/oauth/access_token",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            json={
                "client_id": client_id,
                "client_secret": client_secret,
                "code": code,
            },
        )
        token_data = token_response.json()

    access_token = token_data.get("access_token")
    if not access_token:
        raise HTTPException(status_code=400, detail="Failed to obtain access token")

    github_user_data = await get_github_user_data(access_token)
    user = curd_account.save_oauth_account(
        db=db, user_data=github_user_data, access_token=access_token
    )

    full_redirect_url = coreutils.abs_url(req, state)
    response = RedirectResponse(url=full_redirect_url)

    access_token_expires = timedelta(minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = Token(
        access_token=security.create_access_token(
            user.id, expires_delta=access_token_expires
        )
    )

    response.set_cookie(
        key=settings.COOKIE_ACCESS_TOKEN,
        value=access_token.access_token,
        httponly=True,
        secure=True,  # 在生产环境中建议开启
        samesite="Lax",  # 防止 CSRF 攻击
    )
    return response


@router.get("/auth/github/authorize", include_in_schema=False)
async def github_oauth_authorize(req: Request):
    client_id = settings.GITHUB_CLIENT_ID
    if not client_id:
        raise HTTPException(
            status_code=500, detail="Missing environment variable: GITHUB_CLIENT_ID"
        )

    next_url = req.query_params.get("next", "/")

    callback_path = f"{settings.API_V1_STR}/auth/github/callback"
    redirect_uri = coreutils.abs_url(req, callback_path)

    github_auth_url = "https://github.com/login/oauth/authorize?" + urlencode(
        {"client_id": client_id, "redirect_uri": redirect_uri, "state": next_url}
    )

    return RedirectResponse(github_auth_url)


@router.get("/get_org_api_token", tags=["myhelper"])
async def get_org_api_token():
    organization = await DATABASE.create_organization("org1")
    LOG.info(f"Created organization: {organization}")
    return await create_org_api_token(organization.organization_id)
    return await create_org_api_token(organization.organization_id)



from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from mtmai.models.models import SysItem


async def get_sys_items(session: AsyncSession, type: str):
    result = await session.exec(select(SysItem).where(SysItem.type == type))
    return result.all()



from textwrap import dedent

import httpx
import structlog
from langchain_community.utilities.duckduckgo_search import DuckDuckGoSearchAPIWrapper
from langchain_core.tools import tool
from mtmai.core.config import settings

LOG = structlog.get_logger()


@tool("search_engine")
def search_engine(query: str, results_limit: int = 3, search_engine: str = "ddg"):
    """Search engine to the internet, 可选搜索引擎名称: ddg | searxng"""

    LOG.info(f"调用搜索:{search_engine},limit: {results_limit} {query}")

    if search_engine == "searxng":
        searxng_url = f"{settings.SEARXNG_URL_BASE}/search"
        LOG.info("调用 search ( %s ), %s", searxng_url, query)

        with httpx.Client() as client:
            params = {"q": query, "format": "json"}
            r = client.get(searxng_url, params=params)
            r.raise_for_status()

            search_results = r.json()

            result_list2 = search_results.get("results", [])[:results_limit]

            result_list3 = [
                {
                    "title": x.get("title", ""),
                    "url": x.get("url", ""),
                    "content": x.get("content", ""),
                }
                for x in result_list2
            ]

            content_lines = ["搜索结果:"]
            for x in result_list3:
                content_lines.append(
                    dedent(f"""title: {x.get("title")}
                        content: {x.get("content")}
                        """)
                )

            return (
                "\n".join(content_lines),
                {
                    "artifaceType": "ArtifactSearchResults",
                    "props": {
                        "title": f"{query}的搜索结果",
                        "results": result_list3,
                        "suggestions": search_results.get("suggestions", []),
                        # "infoboxes": search_results.get("infoboxes", []),
                    },
                },
            )

    search_engine = DuckDuckGoSearchAPIWrapper()
    results = search_engine._ddgs_text(query)
    return [{"content": r["body"], "url": r["href"]} for r in results]
    return [{"content": r["body"], "url": r["href"]} for r in results]
